@@ -56,6 +56,8 @@ struct _GisAssistantPrivate
   GtkWidget *forward;
   GtkWidget *back;
   GtkWidget *main_layout;
+  GtkWidget *action_area;
+  GtkWidget *page_action_widget_area;
 
   GList *pages;
   GisPage *current_page;
@@ -82,9 +84,11 @@ widget_destroyed (GtkWidget    *widget,
 }
 
 static void
-gis_assistant_switch_to (GisAssistant *assistant, GisPage *page)
+gis_assistant_switch_to (GisAssistant          *assistant,
+                         GisAssistantDirection  direction,
+                         GisPage               *page)
 {
-  GIS_ASSISTANT_GET_CLASS (assistant)->switch_to (assistant, page);
+  GIS_ASSISTANT_GET_CLASS (assistant)->switch_to (assistant, direction, page);
 }
 
 void
@@ -115,7 +119,7 @@ static void
 gis_assistant_real_next_page (GisAssistant *assistant,
                               GisPage      *page)
 {
-  gis_assistant_switch_to (assistant, find_next_page (page));
+  gis_assistant_switch_to (assistant, GIS_ASSISTANT_NEXT, find_next_page (page));
 }
 
 static GisPage *
@@ -133,7 +137,30 @@ gis_assistant_previous_page (GisAssistant *assistant)
 {
   GisAssistantPrivate *priv = assistant->priv;
   g_return_if_fail (priv->current_page != NULL);
-  gis_assistant_switch_to (assistant, find_prev_page (priv->current_page));
+  gis_assistant_switch_to (assistant, GIS_ASSISTANT_PREV, find_prev_page (priv->current_page));
+}
+
+static void
+remove_from_page_action_area (GtkWidget *widget,
+                              gpointer   user_data)
+{
+  GisAssistantPrivate *priv = user_data;
+  gtk_container_remove (GTK_CONTAINER (priv->page_action_widget_area), widget);
+}
+
+static void
+update_action_widget (GisAssistant *assistant)
+{
+  GisAssistantPrivate *priv = assistant->priv;
+  GtkWidget *action;
+
+  gtk_container_foreach (GTK_CONTAINER (priv->page_action_widget_area),
+                         remove_from_page_action_area, priv);
+
+  action = gis_page_get_action_widget (priv->current_page);
+  if (action)
+    gtk_container_add (GTK_CONTAINER (priv->page_action_widget_area),
+                       action);
 }
 
 static void
@@ -149,19 +176,19 @@ update_navigation_buttons (GisAssistant *assistant)
 
   page_priv = page->assistant_priv;
 
-  can_go_forward = gis_page_get_complete (page);
-  gtk_widget_set_sensitive (priv->forward, can_go_forward);
-
   is_first_page = (page_priv->link->prev == NULL);
+  gtk_widget_set_sensitive (priv->back, !is_first_page);
+
   is_last_page = (page_priv->link->next == NULL);
-  gtk_widget_set_visible (priv->back, !is_first_page);
-  gtk_widget_set_visible (priv->forward, !is_last_page);
+  can_go_forward = gis_page_get_complete (page) && !is_last_page;
+  gtk_widget_set_sensitive (priv->forward, can_go_forward);
 }
 
 static void
 gis_assistant_real_prepare (GisAssistant *assistant,
                             GisPage      *page)
 {
+  update_action_widget (assistant);
   update_navigation_buttons (assistant);
 }
 
@@ -201,24 +228,18 @@ gis_assistant_add_page (GisAssistant *assistant,
     update_navigation_buttons (assistant);
 }
 
-void
-gis_assistant_destroy_all_pages (GisAssistant *assistant)
+GisPage *
+gis_assistant_get_current_page (GisAssistant *assistant)
 {
   GisAssistantPrivate *priv = assistant->priv;
-  GList *l, *next;
+  return priv->current_page;
+}
 
-  g_object_freeze_notify (G_OBJECT (assistant));
-
-  for (l = priv->pages; l != NULL; l = next)
-    {
-      GisPage *page = l->data;
-      next = l->next;
-      gtk_widget_destroy (GTK_WIDGET (page));
-    }
-
-  g_object_thaw_notify (G_OBJECT (assistant));
-
-  g_assert (priv->pages == NULL);
+GList *
+gis_assistant_get_all_pages (GisAssistant *assistant)
+{
+  GisAssistantPrivate *priv = assistant->priv;
+  return priv->pages;
 }
 
 static void
@@ -268,27 +289,40 @@ void
 gis_assistant_locale_changed (GisAssistant *assistant)
 {
   GisAssistantPrivate *priv = assistant->priv;
+  GList *l;
+
   gtk_button_set_label (GTK_BUTTON (priv->forward), _("_Next"));
   gtk_button_set_label (GTK_BUTTON (priv->back), _("_Back"));
+
+  for (l = priv->pages; l != NULL; l = l->next)
+    gis_page_locale_changed (l->data);
 }
 
 static void
 gis_assistant_init (GisAssistant *assistant)
 {
   GisAssistantPrivate *priv = GET_PRIVATE (assistant);
-  GtkWidget *action_area;
+  GtkWidget *navigation;
   assistant->priv = priv;
 
-  priv->main_layout = gtk_box_new (GTK_ORIENTATION_VERTICAL, 10);
+  priv->main_layout = gtk_box_new (GTK_ORIENTATION_VERTICAL, 20);
   gtk_box_pack_start (GTK_BOX (assistant), priv->main_layout, TRUE, TRUE, 0);
 
   priv->frame = gtk_frame_new ("");
   gtk_frame_set_shadow_type (GTK_FRAME (priv->frame), GTK_SHADOW_NONE);
   gtk_box_pack_start (GTK_BOX (priv->main_layout), priv->frame, TRUE, TRUE, 0);
 
-  action_area = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 6);
-  gtk_box_set_homogeneous (GTK_BOX (action_area), TRUE);
-  gtk_widget_set_halign (action_area, GTK_ALIGN_END);
+  priv->action_area = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 6);
+
+  priv->page_action_widget_area = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
+  gtk_box_pack_start (GTK_BOX (priv->action_area), priv->page_action_widget_area,
+                      FALSE, FALSE, 0);
+
+  navigation = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 6);
+  gtk_box_set_homogeneous (GTK_BOX (navigation), TRUE);
+  gtk_widget_set_halign (navigation, GTK_ALIGN_END);
+  gtk_widget_set_hexpand (navigation, TRUE);
+  gtk_box_pack_end (GTK_BOX (priv->action_area), navigation, TRUE, TRUE, 0);
 
   priv->forward = gtk_button_new ();
   gtk_button_set_image (GTK_BUTTON (priv->forward),
@@ -301,14 +335,14 @@ gis_assistant_init (GisAssistant *assistant)
                         gtk_image_new_from_stock (GTK_STOCK_GO_BACK, GTK_ICON_SIZE_BUTTON));
   gtk_button_set_use_underline (GTK_BUTTON (priv->back), TRUE);
 
-  gtk_box_pack_start (GTK_BOX (action_area), priv->back, FALSE, FALSE, 0);
-  gtk_box_pack_start (GTK_BOX (action_area), priv->forward, FALSE, FALSE, 0);
+  gtk_box_pack_start (GTK_BOX (navigation), priv->back, FALSE, FALSE, 0);
+  gtk_box_pack_start (GTK_BOX (navigation), priv->forward, FALSE, FALSE, 0);
 
   g_signal_connect (priv->forward, "clicked", G_CALLBACK (go_forward), assistant);
   g_signal_connect (priv->back, "clicked", G_CALLBACK (go_backward), assistant);
 
   gis_assistant_locale_changed (assistant);
-  gtk_box_pack_start (GTK_BOX (priv->main_layout), action_area, FALSE, TRUE, 0);
+  gtk_box_pack_start (GTK_BOX (priv->main_layout), priv->action_area, FALSE, TRUE, 0);
 
   gtk_widget_show_all (GTK_WIDGET (assistant));
 }
@@ -351,6 +385,8 @@ gis_assistant_class_init (GisAssistantClass *klass)
                          NULL,
                          G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
 
+  g_object_class_install_properties (gobject_class, PROP_LAST, obj_props);
+
   /**
    * GisAssistant::next-page:
    * @assistant: the #GisAssistant
