@@ -26,6 +26,7 @@
 #define PAGE_ID "account"
 
 #include "config.h"
+#include "account-resources.h"
 #include "gis-account-page.h"
 
 #include <glib/gi18n.h>
@@ -61,7 +62,9 @@ struct _GisAccountPagePrivate
   gboolean valid_name;
   gboolean valid_username;
   gboolean valid_password;
+  gboolean valid_confirm;
   const gchar *password_reason;
+  guint reason_timeout;
   ActUserAccountType account_type;
 
   gboolean user_data_unsaved;
@@ -70,6 +73,8 @@ struct _GisAccountPagePrivate
   guint realmd_watch;
   UmRealmManager *realm_manager;
   gboolean domain_chosen;
+
+  GtkWidget *action;
 };
 
 #define OBJ(type,name) ((type)gtk_builder_get_object(GIS_PAGE (page)->builder,(name)))
@@ -115,6 +120,7 @@ clear_account_page (GisAccountPage *page)
   priv->valid_name = FALSE;
   priv->valid_username = FALSE;
   priv->valid_password = FALSE;
+  priv->valid_confirm = FALSE;
 
   /* FIXME: change this for a large deployment scenario; maybe through a GSetting? */
   priv->account_type = ACT_USER_ACCOUNT_TYPE_ADMINISTRATOR;
@@ -131,7 +137,10 @@ static gboolean
 local_validate (GisAccountPage *page)
 {
   GisAccountPagePrivate *priv = page->priv;
-  return priv->valid_name && priv->valid_username && priv->valid_password;
+  return priv->valid_name &&
+         priv->valid_username &&
+         priv->valid_password &&
+         priv->valid_confirm;
 }
 
 static gboolean
@@ -205,8 +214,6 @@ set_has_enterprise (GisAccountPage *page,
     return;
 
   priv->has_enterprise = has_enterprise;
-  gtk_widget_set_visible (WID ("local-button"), has_enterprise);
-  gtk_widget_set_visible (WID ("enterprise-button"), has_enterprise);
 
   if (!has_enterprise)
     set_mode (page, UM_LOCAL);
@@ -273,7 +280,42 @@ username_changed (GtkComboBoxText *combo,
   update_account_page_status (page);
 }
 
-#define MIN_PASSWORD_LEN 6
+static gboolean
+reason_timeout_cb (gpointer data)
+{
+  GisAccountPage *page = GIS_ACCOUNT_PAGE (data);
+  GisAccountPagePrivate *priv = page->priv;
+  GtkWidget *password_entry;
+  GtkWidget *confirm_entry;
+  const gchar *password;
+  const gchar *verify;
+
+  password_entry = WID("account-password-entry");
+  confirm_entry = WID("account-confirm-entry");
+  password = gtk_entry_get_text (GTK_ENTRY (password_entry));
+  verify = gtk_entry_get_text (GTK_ENTRY (confirm_entry));
+
+  if (strlen (password) > 0 && !priv->valid_password)
+    set_entry_validation_error (GTK_ENTRY (password_entry), priv->password_reason);
+
+  if (strlen (verify) > 0 && !priv->valid_confirm)
+    set_entry_validation_error (GTK_ENTRY (confirm_entry), _("Passwords do not match"));
+
+  priv->reason_timeout = 0;
+
+  return G_SOURCE_REMOVE;
+}
+
+static void
+refresh_reason_timeout (GisAccountPage *page)
+{
+  GisAccountPagePrivate *priv = page->priv;
+
+  if (page->priv->reason_timeout != 0)
+    g_source_remove (page->priv->reason_timeout);
+
+  page->priv->reason_timeout = g_timeout_add (600, reason_timeout_cb, page);
+}
 
 static void
 update_password_entries (GisAccountPage *page)
@@ -286,15 +328,18 @@ update_password_entries (GisAccountPage *page)
   GtkWidget *confirm_entry;
   GtkWidget *username_combo;
   GtkWidget *password_strength;
+  GtkWidget *strength_label;
   gdouble strength;
   gint strength_level;
   const gchar *hint;
   const gchar *long_hint = NULL;
+  gchar *strength_hint;
 
   password_entry = WID("account-password-entry");
   confirm_entry = WID("account-confirm-entry");
   username_combo = WID("account-username-combo");
   password_strength = WID("account-password-strength");
+  strength_label = WID("account-password-strength-label");
 
   password = gtk_entry_get_text (GTK_ENTRY (password_entry));
   verify = gtk_entry_get_text (GTK_ENTRY (confirm_entry));
@@ -303,69 +348,83 @@ update_password_entries (GisAccountPage *page)
   strength = pw_strength (password, NULL, username, &hint, &long_hint, &strength_level);
   gtk_level_bar_set_value (GTK_LEVEL_BAR (password_strength), strength_level);
 
+  if (strlen (password) == 0)
+    strength_hint = g_strdup ("");
+  else
+    strength_hint = g_strdup_printf (_("Strength: %s"), hint);
+  gtk_label_set_label (GTK_LABEL (strength_label), strength_hint);
+  g_free (strength_hint);
+
   if (strength == 0.0) {
     priv->valid_password = FALSE;
     priv->password_reason = long_hint ? long_hint : hint;
   }
-  else if (strcmp (password, verify) != 0) {
-    priv->valid_password = FALSE;
-    priv->password_reason = _("Passwords do not match");
-  }
   else {
     priv->valid_password = TRUE;
+    priv->password_reason = NULL;
   }
+
+  priv->valid_confirm = priv->valid_password &&
+                        strcmp (password, verify) == 0;
+
+  if (priv->valid_password) {
+    clear_entry_validation_error (GTK_ENTRY (password_entry));
+    gtk_widget_set_sensitive (confirm_entry, TRUE);
+  } else {
+    gtk_widget_set_sensitive (confirm_entry, FALSE);
+    gtk_entry_set_text (GTK_ENTRY (confirm_entry), "");
+    clear_entry_validation_error (GTK_ENTRY (confirm_entry));
+  }
+
+  refresh_reason_timeout (page);
 }
 
 static void
 password_changed (GtkWidget      *w,
                   GParamSpec     *pspec,
                   GisAccountPage *page)
-{
-  GisAccountPagePrivate *priv = page->priv;
-  update_password_entries (page);
-
-  priv->user_data_unsaved = TRUE;
-  update_account_page_status (page);
-}
-
-static void
-confirm_changed (GtkWidget      *w,
-                 GParamSpec     *pspec,
-                 GisAccountPage *page)
 {
   GisAccountPagePrivate *priv = page->priv;
   clear_entry_validation_error (GTK_ENTRY (w));
   update_password_entries (page);
-
   priv->user_data_unsaved = TRUE;
   update_account_page_status (page);
 }
 
 static gboolean
-confirm_entry_focus_out (GtkWidget      *entry,
+password_entry_focus_out (GtkWidget      *widget,
+                          GdkEventFocus  *event,
+                          GisAccountPage *page)
+{
+  GisAccountPagePrivate *priv = page->priv;
+  GtkEntry *entry = GTK_ENTRY (widget);
+
+  if (page->priv->reason_timeout != 0)
+    g_source_remove (page->priv->reason_timeout);
+
+  if (!priv->valid_password)
+    set_entry_validation_error (entry, priv->password_reason);
+  else
+    clear_entry_validation_error (entry);
+
+  return FALSE;
+}
+
+static gboolean
+confirm_entry_focus_out (GtkWidget      *widget,
                          GdkEventFocus  *event,
                          GisAccountPage *page)
 {
   GisAccountPagePrivate *priv = page->priv;
-  const gchar *password;
+  GtkEntry *entry = GTK_ENTRY (widget);
   const gchar *verify;
-  GtkEntry *password_entry;
-  GtkEntry *confirm_entry;
 
-  password_entry = OBJ(GtkEntry*, "account-password-entry");
-  confirm_entry = OBJ(GtkEntry*, "account-confirm-entry");
-  password = gtk_entry_get_text (password_entry);
-  verify = gtk_entry_get_text (confirm_entry);
+  verify = gtk_entry_get_text (entry);
 
-  if (strlen (password) > 0 && strlen (verify) > 0) {
-    if (!priv->valid_password) {
-      set_entry_validation_error (confirm_entry,
-                                  priv->password_reason);
-    }
-    else {
-      clear_entry_validation_error (confirm_entry);
-    }
-  }
+  if (strlen (verify) > 0 && !priv->valid_confirm)
+    set_entry_validation_error (entry, _("Passwords do not match"));
+  else
+    clear_entry_validation_error (entry);
 
   return FALSE;
 }
@@ -420,7 +479,8 @@ local_create_user (GisAccountPage *page)
   /* this can happen when going back */
   if (!priv->valid_name ||
       !priv->valid_username ||
-      !priv->valid_password) {
+      !priv->valid_password ||
+      !priv->valid_confirm) {
     return;
   }
 
@@ -813,17 +873,11 @@ on_entry_changed (GtkEditable *editable,
 }
 
 static void
-on_local_clicked (GtkButton *button,
-                  gpointer   user_data)
+toggle_mode (GtkToggleButton *button,
+             gpointer         user_data)
 {
-  set_mode (GIS_ACCOUNT_PAGE (user_data), UM_LOCAL);
-}
-
-static void
-on_enterprise_clicked (GtkButton *button,
-                       gpointer   user_data)
-{
-  set_mode (GIS_ACCOUNT_PAGE (user_data), UM_ENTERPRISE);
+  set_mode (GIS_ACCOUNT_PAGE (user_data),
+            gtk_toggle_button_get_active (button) ? UM_ENTERPRISE : UM_LOCAL);
 }
 
 static void
@@ -868,7 +922,9 @@ gis_account_page_constructed (GObject *object)
   g_signal_connect (password_entry, "notify::text",
                     G_CALLBACK (password_changed), page);
   g_signal_connect (confirm_entry, "notify::text",
-                    G_CALLBACK (confirm_changed), page);
+                    G_CALLBACK (password_changed), page);
+  g_signal_connect_after (password_entry, "focus-out-event",
+                          G_CALLBACK (password_entry_focus_out), page);
   g_signal_connect_after (confirm_entry, "focus-out-event",
                           G_CALLBACK (confirm_entry_focus_out), page);
 
@@ -876,10 +932,6 @@ gis_account_page_constructed (GObject *object)
                     G_CALLBACK (on_domain_changed), page);
   g_signal_connect (WID("enterprise-login"), "changed",
                     G_CALLBACK (on_entry_changed), page);
-  g_signal_connect (WID("local-button"), "clicked",
-                    G_CALLBACK (on_local_clicked), page);
-  g_signal_connect (WID("enterprise-button"), "clicked",
-                    G_CALLBACK (on_enterprise_clicked), page);
 
   priv->act_client = act_user_manager_get_default ();
 
@@ -890,12 +942,15 @@ gis_account_page_constructed (GObject *object)
 
   priv->has_enterprise = FALSE;
 
+  priv->action = gtk_toggle_button_new_with_mnemonic ("_Use Enterprise Login");
+  g_signal_connect (priv->action, "toggled", G_CALLBACK (toggle_mode), page);
+  gtk_widget_show (priv->action);
+  g_object_ref_sink (priv->action);
+
   /* force a refresh by setting to an invalid value */
   priv->mode = NUM_MODES;
   set_mode (page, UM_LOCAL);
 
-  gis_page_set_title (GIS_PAGE (page), _("Login"));
-
   gtk_widget_show (GTK_WIDGET (page));
 }
 
@@ -908,10 +963,23 @@ gis_account_page_dispose (GObject *object)
   g_clear_object (&priv->act_user);
   g_clear_object (&priv->act_client);
   g_clear_object (&priv->realm_manager);
+  g_clear_object (&priv->action);
 
   G_OBJECT_CLASS (gis_account_page_parent_class)->dispose (object);
 }
 
+static GtkWidget *
+gis_account_page_get_action_widget (GisPage *page)
+{
+  return GIS_ACCOUNT_PAGE (page)->priv->action;
+}
+
+static void
+gis_account_page_locale_changed (GisPage *page)
+{
+  gis_page_set_title (GIS_PAGE (page), _("Login"));
+}
+
 static void
 gis_account_page_class_init (GisAccountPageClass *klass)
 {
@@ -919,6 +987,8 @@ gis_account_page_class_init (GisAccountPageClass *klass)
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
   page_class->page_id = PAGE_ID;
+  page_class->locale_changed = gis_account_page_locale_changed;
+  page_class->get_action_widget = gis_account_page_get_action_widget;
   object_class->constructed = gis_account_page_constructed;
   object_class->dispose = gis_account_page_dispose;
 
@@ -928,6 +998,7 @@ gis_account_page_class_init (GisAccountPageClass *klass)
 static void
 gis_account_page_init (GisAccountPage *page)
 {
+  g_resources_register (account_get_resource ());
   page->priv = GET_PRIVATE (page);
 }
 
